@@ -3,14 +3,46 @@ A function-oriented testing framework for Python 3.
 
 See README.md
 """
+import os
 import time
 import inspect
 import types
+import traceback
+import io
+import re
+import json
+import tempfile
+import shutil
+from contextlib import redirect_stdout, redirect_stderr
+import dataclasses
 from functools import wraps
 from contextlib import contextmanager
 from random import shuffle
 
 
+log_fp = None
+log_last_time = None
+
+
+def log(*args):
+    global log_fp, log_last_time
+    if log_fp is None:
+        log_fp = open("log.txt", "a")
+    delta = 0
+    if log_last_time is not None:
+        delta = time.time() - log_last_time
+    log_last_time = time.time()
+    log_fp.write(f"{delta:3.1f} " + "".join([str(i) + " " for i in args]) + "\n")
+    log_fp.flush()
+
+
+ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
+
+
+def strip_ansi(line):
+    return ansi_escape.sub("", line)
+
+
 def get_class_or_module_that_defined_method(meth):
     # From https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
     if inspect.ismethod(meth):
@@ -44,6 +76,12 @@ class TrappedException(Exception):
     pass
 
 
+class SkipException(Exception):
+    def __init__(self, full_name, reason):
+        self.full_name = full_name
+        self.reason = reason
+
+
 class MockFunction:
     def __init__(self, replacing_func=None):
         if replacing_func is not None:
@@ -162,6 +200,47 @@ class MockFunction:
         return kws == self.normalized_calls()[0]
 
 
+class JSONDataClassEncoder(json.JSONEncoder):
+    def default(self, o):
+        if isinstance(o, BaseException):
+            return f"{o.__class__.__name__}(\"{str(o)}\")"
+        if isinstance(o, ZestResult):
+            if o.error is not None:
+                try:
+                    dataclasses.asdict(o)
+                except Exception as e:
+                    # If it can not be encoded convert to str
+                    o.error = Exception(f"{o.error.__class__.__name__}: \"{str(o.error)}\"")
+        if dataclasses.is_dataclass(o):
+            return dataclasses.asdict(o)
+        return super().default(o)
+
+
+@dataclasses.dataclass
+class ZestResult:
+    call_stack: list
+    full_name: str
+    short_name: str
+    error: Exception = None
+    error_formatted: str = None
+    elapsed: float = None
+    skip: str = None
+    stdout: str = None
+    stderr: str = None
+    source: str = None
+    pid: int = None
+    is_running: bool = False
+    is_starting: bool = False
+    worker_i: int = 0
+
+    def dumps(self):
+        return json.dumps(self, cls=JSONDataClassEncoder)
+
+    @classmethod
+    def loads(cls, s):
+        return ZestResult(**json.loads(s))
+
+
 class zest:
     """
     This is a helper to make calling a little bit cleaner.
@@ -184,6 +263,7 @@ class zest:
     zest(some_test)
     """
 
+    # TODO: Convert these to just use a list of zest_results
     _call_log = []
     _call_stack = []
     _call_errors = []
@@ -194,9 +274,14 @@ class zest:
     _mock_stack = []
     _allow_to_run = None
     _disable_shuffle = False
+    _capture = False
+    _bypass_skip = []
+    _current_error = None   # Useful for checks by _after
+    _common_tmp = None
+    _tmp_root = None
 
     @staticmethod
-    def reset():
+    def reset(disable_shuffle=False, bypass_skip=None, common_tmp=None, tmp_root=None):
         zest._call_log = []
         zest._call_stack = []
         zest._call_errors = []
@@ -206,20 +291,62 @@ class zest:
         zest._test_stop_callback = None
         zest._mock_stack = []
         zest._allow_to_run = None
-        zest._disable_shuffle = False
+        zest._capture = False
+        zest._disable_shuffle = disable_shuffle
+        zest._bypass_skip = [] if bypass_skip is None else bypass_skip.split(":")
+        zest._common_tmp = common_tmp
+        zest._tmp_root = tmp_root
 
     @staticmethod
-    def skip(code="s", reason=None):
+    def current_test_name():
+        return zest._call_stack[-1]
+
+    @staticmethod
+    def current_test_full_name():
+        return ".".join(zest._call_stack)
+
+    @staticmethod
+    def current_test_error():
+        """
+        Current error is a useful state to check in _after()
+        """
+        return zest._current_error
+
+    # TODO: Sort out all the naming conventions for this
+    # @staticmethod
+    # def parameter_list(params_list):
+    #     """
+    #     Params list is a list of tuples that will be passed to the *args.
+    #     If param_list is not a list of tuples they will be converted to tuples
+    #     """
+    #
+    #     def decorator(fn):
+    #         @wraps(fn)
+    #         def wrapper(*params, **kwargs):
+    #             fn(*params, **kwargs)
+    #
+    #         _params_list = [
+    #             params if isinstance(params, tuple) else (params,)
+    #             for params in params_list
+    #         ]
+    #         setattr(wrapper, "params_list", _params_list)
+    #
+    #         return wrapper
+    #
+    #     return decorator
+
+    @staticmethod
+    def skip(reason=None):
         def decorator(fn):
             @wraps(fn)
             def wrapper(*args, **kwargs):
-                return fn(*args, **kwargs)
-
-            if len(code) > 1:
-                raise ValueError("Skip codes can only be one character")
+                full_name = ".".join(zest._call_stack)
+                if full_name not in zest._bypass_skip:
+                    raise SkipException(full_name, reason)
+                else:
+                    fn(*args, **kwargs)
 
             setattr(wrapper, "skip", True)
-            setattr(wrapper, "skip_code", code)
             setattr(wrapper, "skip_reason", reason)
             return wrapper
 
@@ -237,6 +364,23 @@ class zest:
 
         return decorator
 
+    @staticmethod
+    def retry(n_tries):
+        def decorator(fn):
+            @wraps(fn)
+            def wrapper(*args, **kwargs):
+                tries = n_tries
+                while tries > 0:
+                    try:
+                        return fn(*args, **kwargs)
+                    except Exception as e:
+                        tries -= 1
+                        if tries == 0:
+                            raise e
+            return wrapper
+
+        return decorator
+
     @staticmethod
     def _setup_mock(symbol, substitute_fn=None):
         if not callable(symbol):
@@ -308,6 +452,12 @@ class zest:
         """
         Use this in the inner most test. Do not attempt to encapsulate
         more than one test with this context. See README.
+
+        The kwargs can include statements about the exception
+
+            in_args=value
+            in_{key}=substring
+            key=val
         """
         got_expected_exception = False
         trapped_exception = TrappedException()
@@ -350,7 +500,13 @@ class zest:
         # If some other exception was raised that should just bubble as usual
 
     @staticmethod
-    def do(*funcs, test_start_callback=None, test_stop_callback=None):
+    def current_test_name():
+        return zest._call_stack[-1]
+
+    @staticmethod
+    def do(
+        *funcs, test_start_callback=None, test_stop_callback=None, allow_to_run=None
+    ):
         """
         This is the entrypoint of any zest at any depth.
 
@@ -382,7 +538,7 @@ class zest:
         Call _before() (if defined) before each test
         Call _after() (if defined) after each test
 
-        The class member _allow_to_run potentialy contains a list of
+        The class member _allow_to_run potentially contains a list of
         zests that are allowed to execute in dotted form. Eg using above:
             ["zest_test1.it_does_y.it_does_y1"]
 
@@ -393,17 +549,21 @@ class zest:
         Eg: ["zest_test1.it_does_y"] means that it_does_y1 will run too.
 
         """
+
         prev_test_start_callback = None
         prev_test_stop_callback = None
+        prev_allow_to_run = None
         if test_start_callback is not None:
             prev_test_start_callback = zest._test_start_callback
             zest._test_start_callback = test_start_callback
         if test_stop_callback is not None:
             prev_test_stop_callback = zest._test_stop_callback
             zest._test_stop_callback = test_stop_callback
+        if allow_to_run is not None:
+            prev_allow_to_run = zest._allow_to_run
+            zest._allow_to_run = allow_to_run
 
         try:
-
             callers_special_local_funcs = {}
 
             if len(funcs) > 0:
@@ -454,73 +614,158 @@ class zest:
                     "A _begin function was declared. Maybe you meant _before?"
                 )
 
+            def _run(name, func, so, se):
+                # params_list = [tuple()]
+                # if hasattr(func, "params_list"):
+                #     params_list = getattr(func, "params_list")
+
+                zest._call_stack += [name]
+                zest._current_error = None
+
+                try:
+                    full_name = ".".join(zest._call_stack)
+                    if (
+                        zest._allow_to_run is not None
+                        and full_name not in zest._allow_to_run
+                        and zest._allow_to_run != "__all__"
+                    ):
+                        zest._call_stack.pop()
+                        return
+
+                except Exception as e:
+                    log(f"EXCEPTION during allow to check run. NAME {name} e {e}")
+                    zest._call_stack.pop()
+                    return
+
+                pre_cwd = os.getcwd()
+                remove_tmp_dir = None
+                try:
+                    if zest._common_tmp is not None:
+                        cwd = zest._common_tmp
+                    else:
+                        # Create a tmp folder per test
+                        tmp_root = zest._tmp_root or "/tmp"
+                        cwd = tempfile.mkdtemp(dir=tmp_root)
+                        remove_tmp_dir = cwd
+
+                    # Set each test into the correct tmp folder
+                    os.chdir(cwd)
+
+                    # for params in params_list:
+                    _before = callers_special_local_funcs.get("_before")
+                    if _before:
+                        try:
+                            _before()
+                        except Exception as e:
+                            zest._call_errors += [(e, zest._call_stack.copy())]
+                            s = (
+                                f"There was an exception while running '_before()' in test '{name}'. "
+                                f"This may mean that the sub-tests are not enumerated and therefore can not be run."
+                            )
+                            zest._call_warnings += [s]
+
+                    try:
+                        zest._call_tree += [full_name]
+                        zest._call_log += [full_name]
+
+                        if zest._test_start_callback:
+                            zest._test_start_callback(
+                                ZestResult(
+                                    zest._call_stack,
+                                    full_name,
+                                    zest._call_stack[-1],
+                                    None,
+                                    None,
+                                    None,
+                                    None,
+                                    None,
+                                    None,
+                                    func.__code__.co_filename,
+                                    os.getpid(),
+                                    True,
+                                )
+                            )
+
+                        error = None
+                        error_formatted = None
+                        skip_reason = None
+                        start_time = time.time()
+                        try:
+                            zest._mock_stack += [[]]
+
+                            try:
+                                func()
+                            except SkipException as e:
+                                skip_reason = e.reason
+                            zest._clear_stack_mocks()
+                            zest._mock_stack.pop()
+                        except Exception as e:
+                            error = e
+                            error_formatted = traceback.format_exception(
+                                etype=type(error), value=error, tb=error.__traceback__
+                            )
+                            zest._call_errors += [
+                                (e, error_formatted, zest._call_stack.copy())
+                            ]
+                            zest._current_error = e
+                        finally:
+                            stop_time = time.time()
+                            if zest._test_stop_callback:
+                                zest_result = ZestResult(
+                                    zest._call_stack,
+                                    ".".join(zest._call_stack),
+                                    zest._call_stack[-1],
+                                    error,
+                                    error_formatted,
+                                    stop_time - start_time,
+                                    skip_reason,
+                                    so.getvalue() if so is not None else None,
+                                    se.getvalue() if se is not None else None,
+                                    func.__code__.co_filename,
+                                    os.getpid(),
+                                    False,
+                                )
+                                zest._test_stop_callback(zest_result)
+
+                        _after = callers_special_local_funcs.get("_after")
+                        if _after:
+                            _after()
+                    except Exception as e:
+                        log(f"ZEST EXCEPTION 1 {e}")
+                    finally:
+                        zest._call_stack.pop()
+                finally:
+                    # Clean up tmp folders if needed
+                    if remove_tmp_dir:
+                        try:
+                            shutil.rmtree(remove_tmp_dir)
+                        except OSError as exc:
+                            if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
+                                raise  # re-raise exception
+                    os.chdir(pre_cwd)
+
             for name, func in funcs_to_call:
                 if len(zest._mock_stack) > 0:
                     for mock_tuple in zest._mock_stack[-1]:
                         if mock_tuple[4]:  # if reset_before_each is set
                             mock_tuple[3].reset()  # Tell the mock to reset
 
-                _before = callers_special_local_funcs.get("_before")
-                if _before:
-                    try:
-                        _before()
-                    except Exception as e:
-                        zest._call_errors += [(e, zest._call_stack.copy())]
-                        s = (
-                            f"There was an exception while running '_before()' in test '{name}'. "
-                            f"This may mean that the sub-tests are not enumerated and therefore can not be run."
-                        )
-                        zest._call_warnings += [s]
+                if zest._capture:
+                    so = io.StringIO()
+                    se = io.StringIO()
+                    with redirect_stdout(so):
+                        with redirect_stderr(se):
+                            _run(name, func, so, se)
+                else:
+                    _run(name, func, None, None)
 
-                zest._call_stack += [name]
-
-                full_name = ".".join(zest._call_stack)
-                if (
-                    zest._allow_to_run is not None
-                    and full_name not in zest._allow_to_run
-                ):
-                    zest._call_stack.pop()
-                    continue
-
-                zest._call_tree += [full_name]
-                zest._call_log += [name]
-
-                if zest._test_start_callback:
-                    zest._test_start_callback(
-                        name, call_stack=zest._call_stack, func=func
-                    )
-
-                error = None
-                start_time = time.time()
-                try:
-                    if not hasattr(func, "skip"):
-                        zest._mock_stack += [[]]
-                        func()
-                        zest._clear_stack_mocks()
-                        zest._mock_stack.pop()
-                except Exception as e:
-                    error = e
-                    zest._call_errors += [(e, zest._call_stack.copy())]
-                finally:
-                    stop_time = time.time()
-                    if zest._test_stop_callback:
-                        zest._test_stop_callback(
-                            name,
-                            call_stack=zest._call_stack,
-                            error=error,
-                            elapsed=stop_time - start_time,
-                            func=func,
-                        )
-                    zest._call_stack.pop()
-
-                _after = callers_special_local_funcs.get("_after")
-                if _after:
-                    _after()
         finally:
             if prev_test_start_callback is not None:
                 zest._test_start_callback = prev_test_start_callback
             if prev_test_stop_callback is not None:
                 zest._test_stop_callback = prev_test_stop_callback
+            if prev_allow_to_run is not None:
+                zest._allow_to_run = prev_allow_to_run
 
     def __init__(self, *args, **kwargs):
         self.do(*args, **kwargs)
